#!/usr/bin/env python
############################################################################
# Noah J Edelson, LBNL
# See Copyright for copyright notice!
###########################################################################


from ConfigParser import *
import os, sys, time, popen2, select, exceptions

from vipus.dag.condor.condorxml import ParseDAGManXMLLog, tailfileobject_condorxml

from vipus.dag.dag import DAG
from vipus.dag.condor.condor_job import Job


def parse_dagman(filename):
    f = open(filename)
    joblist = []
    for line in f.readline():
        if line == '':
            break
        if line.find('Job ') > -1:           # found 'Job'
            jobfname = line.split(' ')[2]
            job = Job(None)
            job.set_filename(os.path.join(dag.submitDir, jobfname))
            job.readFile(dagfilename.split('/')[0] + os.sep + jobfname)
            joblist += [job]
        if line.find('PARENT ') > -1:        # found 'PARENT'
            parentID = str(line.split(' ')[1])
            childIDs = (line.split(' ')[3:])  # find all the children
            kidlist = []
            for jobber in joblist:            
                if jobber.get_filename().find(parentID) > -1:  # assign parent
                    for potentialKid in joblist:
                        if jobber != potentialKid:      
                            jobber.attachDependency( [ potentialKid ] )
    return joblist

def write_dagman(filename, submitDir, joblist):
    dagfile = open(filename, 'w')
    for job in joblist:
        job.writeFile( os.path.join(submitDir, job.jobfilename))
        dagfile.write('Job ' + job.joblabel + ' ' + job.jobfilename + '\n')
        if job.pre:   dagfile.write('SCRIPT PRE '  + job.jobfilename + ' ' + job.pre + '\n')
        if job.post:  dagfile.write('SCRIPT POST ' + job.jobfilename + ' ' + job.post + '\n')
        if job.retry: dagfile.write('Retry '       + job.jobfilename + ' ' + str(job.retry) + '\n')
    for job in joblist:
        for dep in job.deplist:
            dagfile.write('PARENT ' + dep.joblabel + ' CHILD ' + job.joblabel + '\n')
    dagfile.close()
    
class CommandFailed(exceptions.Exception): pass            
## time and run a command which could produce errors on stderr with a non-zero error code

def run_command(cmd):
    child=popen2.Popen3(cmd, capturestderr=1)
    stdout = []
    stdout_done = 0
    stderr = []
    stderr_done = 0
       
    ## continuously read from child stdout and stderr until process exists
    while 1:
        fds = select.select([child.fromchild, child.childerr], [], [])
        for d in fds[0]:
            if d == child.fromchild:
                r = child.fromchild.read(1024)
                if r == "":
                    stdout_done = 1
                else:
                    stdout.append(r)
            elif d == child.childerr: 
                r = child.childerr.read(1024)
                if r == "":
                    stderr_done = 1
                else:
                    stderr.append(r)
                   
        ## if both child streams are done, exit (process should already have exited)
        if stdout_done and stderr_done:
            break  
                   
    child_sts = child.wait()
                       
    stdout = "".join(stdout)
    stderr = "".join(stderr)

    if child_sts != 0:
        raise CommandFailed, (child_sts, stdout, stderr)
    return child_sts, stdout, stderr



        
class CondorDAG(DAG):
    """
    Describes a directed acyclic graph, each node being a condor job.
    @param client: 
    """
    def __init__(self, client, filename='default.dag', logfile='default.log'):
        DAG.__init__(self, client, filename, logfile)
        
    def Submit(self):
        """
        Saves the current DAG to disk and calls condor_submit_dag on it.
        """
        DAG.Submit(self)

        curdir = os.getcwd()
        os.chdir(self.submitDir)
        print 'self.filename := ' + str(self.filename)
        try:
            instream, outstream, errstream = run_command('condor_submit_dag -f -log %s -notification Never %s' % (self.logfile, self.filename))
        finally:
            os.chdir(curdir)


    def Watch(self, terminate):
        """
        Watches the current logfile, parsing it atomically as it arrives.  Right?
        @param terminate: The termination event to return on.
        """

        self.BuildJobDict()
        while not os.path.exists(self.full_logfile):
            time.sleep(0.1)
            
        # Open the Condor XML output file
        tfo = tailfileobject_condorxml(self.full_logfile, terminate)

        ## build a dict mapping from job Names to job Objects
        jobDict = {}
        for job in self.joblist:
            jobDict[job.get_label()] = job
        # Create the handler
        dh = ParseDAGManXMLLog(tfo, self.client, jobDict)  #
        dh.parse()
        
    def readFiles(self, dagfilename):  # Job name jobfilename, PARENT parent CHILD childlist0 childlist1 childlist2...
        """
        Reads a .DAG file.        
        @param dagfilename: The .dag filename to populate self's attributes with.
        """
        print 'readFiles: ', dagfilename        
        dag = DAG(None)
        #dag.submitDir = dagfilename.split('/')[]
        joblist = parse_daman(dagfilename)
        dag.addJobs(joblist)
        return dag
    
    def writeFiles(self, dagfilename):
        """
        Writes a .DAG file.
        @param dagfilename: The .dag filename to populate with the self's attributes.
        """
        try: os.mkdir(self.submitDir)
        except: pass
        self.getNames()
        write_dagman(dagfilename, self.submitDir, self.joblist)
        self.dagfilename = dagfilename

def main(argv): print 'vipuslogic library, not for main()ing!'
if __name__ == "__main__" : main(sys.argv)
