### A class which provides a file-like object that continuously reads from a file like the Unix tail command.
### Used for adapting the SAX XML parser to Condor's XML log file format.

import time, os, stat, logging, sys

from vipus.logUtil import logger

class tailfileobject_UnsupportedException:
    pass

class tailfileobject_UnsupportedModeException(tailfileobject_UnsupportedException):
    pass

class tailfileobject_UnsupportedBufferingException(tailfileobject_UnsupportedException):
    pass

class tailfileobject_FileTruncated:
    pass
        
class tailfileobject:
    TIMEOUT = 0.1
    def __init__(self, name, terminate, mode='r', buffering=0):
        self.terminate = terminate
        logger.info("Opening %s" % name)
        if mode != 'r':
            raise tailfileobject_UnsupportedException, "tailfileobject only supports read operations"
        if buffering != 0:
            raise tailfileobject_UnsupportedException, "tailfileobject only supports unbuffered operation"
        self.__file = open(name, "r")
        logger.info("Opened %s" % self.__file)
        self.__fd = self.__file.fileno()
        self.__current_size = os.fstat(self.__fd)[stat.ST_SIZE]
        logger.info("Starting size: %d" % self.__current_size)
        self.__current_position = 0L

    def read(self, size):
        logger.info("Read request (size: %d)" % size)
        if not self.terminate.empty():
            logger.info("Terminate set")
            self.close()
            return ''
        logger.info("Read request: %d bytes" % size)
        new_size = os.fstat(self.__fd)[stat.ST_SIZE]
        logger.info("File size: %d" % new_size)
        if new_size < self.__current_size:
            raise tailfileobject_FileTruncated
        self.__current_size = new_size
        
        ## if there are enough bytes to satisfy this read fully already, return them immediately
        if self.__current_size - self.__current_position >= size:
            logger.info("%d bytes are available (more than size)" % (self.__current_size - self.__current_position))
            data = self.__file.read(size)

        ## if there are bytes available, but less than size, return the bytes available
        elif self.__current_size - self.__current_position:
            logger.info("%d bytes are available (less than size)" % (self.__current_size - self.__current_position))
            data = self.__file.read(self.__current_size - self.__current_position)

        ## otherwise, poll until bytes are available, and return up to size
        else:
            logger.info("Waiting for data")
            while 1:
                logger.debug("Begin while loop")
                if not self.terminate.empty():
                    logger.info("Terminate set")
                    self.close()
                    return ''
                new_size = os.fstat(self.__fd)[stat.ST_SIZE]
                logger.debug("Got size: %d" % new_size)
                if new_size < self.__current_size:
                    logger.info("File truncated")
                    raise tailfileobject_FileTruncated
                self.__current_size = new_size
                if self.__current_size > self.__current_position:
                    logger.info("%d new bytes available" % (self.__current_size - self.__current_position))
                    if self.__current_size - self.__current_position < size:
                        data = self.__file.read(self.__current_size - self.__current_position)
                    else:
                        data = self.__file.read(size)
                    break
                logger.debug("No new data")
                time.sleep(tailfileobject.TIMEOUT)
        logger.info("Returning %d bytes of data" % len(data))
        self.__current_position += len(data)
        return data

    def close(self):
        ## Should we throw an exception if there is unread data?
        self.__file.close()

        
if __name__ == '__main__':
    
    f = tailfileobject("foo")
    while 1:
        print "Data:", f.read(1024)
        
